/*
	SOS Emulator Add-on template
	(c) Mikhail Panasyuk aka BeAR, 2009. E-mail: otoko@yandex.ru

	Contains portions of NewOS kernel code:

	Copyright 2001-2002, Travis Geiselbrecht. All rights reserved. 
	Distributed under the terms of the NewOS License. 
*/

#include "../../sos_emulator_addon.h"

const char* add_on_name = "BeOSrt";

// The run queue. Holds the threads ready to run ordered by priority.
static struct thread_queue gRunQueue = {NULL, NULL};

// The current thread being run.
static struct thread *_currThread = NULL;

// fake functions
struct thread* thread_get_current_thread(void)
{
	if (!_currThread)
		_currThread = gRunQueue.head;

	return _currThread;
}
void context_switch(struct thread *oldThread, struct thread *nextThread)
{
	_currThread = nextThread;
}

static int _rand(void)
{
   static int next = 0;
   if (next == 0) next = system_time();
   next = next * 1103515245 + 12345;
   return((next >> 16) & 0x7FFF);
}

void scheduler_enqueue_in_run_queue(struct thread *thread)
{
	struct thread *curr, *prev;

	for(curr = gRunQueue.head, prev = NULL; 
		curr && (curr->priority >= thread->priority); 
		curr = curr->queue_next) 
	{
		if (prev)
			prev = prev->queue_next;
		else
			prev = gRunQueue.head;
	}

	thread->queue_next = curr;

	if(prev)
		prev->queue_next = thread;
	else
		gRunQueue.head = thread;

}

void scheduler_remove_from_run_queue(struct thread *thread)
{
   struct thread *item, *prev;

   // find thread in run queue
   for (item = gRunQueue.head, prev = NULL; 
        item && item != thread; 
        item = item->queue_next) 
   {
      if (prev)
         prev = prev->queue_next;
      else
         prev = gRunQueue.head;
   }
   
   if (prev)
      prev->queue_next = item->queue_next;
   else
      gRunQueue.head = item->queue_next;
}

/* Skip gaps -- Begin */

#define REAL_TIME_SKIP_GAP 20
#define MIN_REAL_TIME_SKIP_GAP 0
#define MAX_REAL_TIME_SKIP_GAP 20

#define SKIP_GAP 99
#define MIN_SKIP_GAP 0
#define MAX_SKIP_GAP 99

#define IS_IN_RANGE(i,min,max) (i >= min && i <= max)

typedef struct scheduler_skip_gap_info {
	uint32	real_time;
	uint32	normal;
} scheduler_skip_gap_info;

static scheduler_skip_gap_info gSkipGaps = { REAL_TIME_SKIP_GAP, SKIP_GAP };

/* Skip gaps -- End */

scheduler_skip_gap_info get_skip_gaps()
{
	return gSkipGaps;
}

status_t set_skip_gaps(scheduler_skip_gap_info skip_gaps)
{
	if (IS_IN_RANGE(skip_gaps.real_time, MIN_REAL_TIME_SKIP_GAP, MAX_REAL_TIME_SKIP_GAP) &&
		IS_IN_RANGE(skip_gaps.normal, MIN_SKIP_GAP, MAX_SKIP_GAP))
	{
		gSkipGaps = skip_gaps;
		return B_OK;
	}
	return B_ERROR;
}

struct thread* select_next_thread_to_run()
{
	struct thread *nextThread, *nextLower;
	uint32 n;

	nextThread = gRunQueue.head;

	// in case we don't have any non-idle threads to run
	if (nextThread->priority == B_IDLE_PRIORITY) {
		return nextThread;
	}

	// select only from real-time threads if any
	if (nextThread->priority >= B_FIRST_REAL_TIME_PRIORITY) {

		while (1) {
			// find thread with next lower priority
			nextLower = nextThread->queue_next;
			while (nextLower->priority >= nextThread->priority)
				nextLower = nextLower->queue_next;
	
	     	// never skip last real-time subqueue of threads
	     	if (nextLower->priority < B_FIRST_REAL_TIME_PRIORITY)
				break;
			
			n = (uint32)(nextThread->priority - nextLower->priority);
			// never skip too much at once
			if (n > gSkipGaps.real_time) break;

			// skip subqueue of real-time threads sometimes
			if (_rand()%(n+1)) break;
	
			nextThread = nextLower;
		}
		return nextThread;
		
	}

	// select from normal threads
	while (1) {
		// find thread with next lower priority
		nextLower = nextThread->queue_next;
		while (nextLower->priority >= nextThread->priority)
			nextLower = nextLower->queue_next;

     	// never skip last non-idle normal subqueue of threads
     	if (nextLower->priority == B_IDLE_PRIORITY)
			break;

		n = (uint32)(nextThread->priority - nextLower->priority);
		// never skip too much at once
		if (n > gSkipGaps.normal) break;

		// skip subqueue of normal threads sometimes
		if (_rand()%(n+1)) break;

		nextThread = nextLower;
	}
	return nextThread;
}

void scheduler_reschedule(void)
{
  struct thread *oldThread = thread_get_current_thread();
  struct thread *nextThread;

  switch (oldThread->next_state) 
  {
     case B_THREAD_RUNNING:
     case B_THREAD_READY:
        scheduler_enqueue_in_run_queue(oldThread);
        break;

/*     case B_THREAD_SUSPENDED:
        break;

     case THREAD_STATE_FREE_ON_RESCHED:
        // This will hopefully be eliminated once the slab
        // allocator is done
        thread_enqueue(oldThread, &dead_q);
        break;
*/
     default:
        break;
  }

  oldThread->state = oldThread->next_state;

  // select next thread from the run queue
  nextThread = select_next_thread_to_run();

  // extract selected thread from the run queue
  scheduler_remove_from_run_queue(nextThread);
  
  nextThread->state = B_THREAD_RUNNING;
  nextThread->next_state = B_THREAD_READY;

  if (nextThread != oldThread/* || oldThread->cpu->info.preempted*/) 
  {

/*     bigtime_t quantum = 3000; // ToDo: calculate quantum!
     timer *quantum_timer= &oldThread->cpu->info.quantum_timer;

     if (!oldThread->cpu->info.preempted)
        _local_timer_cancel_event(oldThread->cpu->info.cpu_num, quantum_timer);
     
     oldThread->cpu->info.preempted = 0;
     add_timer(quantum_timer, &reschedule_event, quantum, B_ONE_SHOT_RELATIVE_TIMER);
*/
     if (nextThread != oldThread)
        context_switch(oldThread, nextThread);
  }
}
