/*
	SOS Emulator - [S]imple [O]S [S]cheduler Emulator
	(c) Mikhail Panasyuk aka BeAR, 2009. E-mail: otoko@yandex.ru

	Contains portions of NewOS kernel code:

	Copyright 2001-2002, Travis Geiselbrecht. All rights reserved. 
	Distributed under the terms of the NewOS License. 
	

	Original concept and results' interpretation:
	
		SOS Emulator emulates workflow of system scheduler and runs
		SOS Tester-like tests to see how this scheduler performs.
		
		See SOS Tester's descriprion (in sos_tester.cpp) for details.
		
		Schedulers are implemented as add-ons. For more info see example addon(s)
		in /add-ons folder.
	
	
	Use	at your own risk, this program may cause system crash or even burn your
	CPU(s), no warranty is provided whatsoever.
	
	Compile like this: gcc sos_emulator.cpp -lbe -lroot -o sos_emulator
	Run like this: sos_emulator 6000 60 30 99 NewOS >sos_emulator_output.txt
		or simply: sos_emulator
	
	Have fun! ;)
*/


#include <OS.h>
#include <String.h>
#include <List.h>
#include <image.h>

#include <Entry.h>
#include <Directory.h>
#include <Path.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "sos_sched.h"

/* Defaults section -- Begin */

// length of lifecycle
// milliseconds
#define DEATH_TIMER_INT32 1000 // 1 seconds
#define MIN_DEATH_TIMER_INT32 1 // 0.001 second
#define MAX_DEATH_TIMER_INT32 3600000 // 1 hour

// threads' number regulations
#define NUM_THREADS 20
#define MIN_NUM_THREADS 2
#define MAX_NUM_THREADS 4096

// threads' priority regulations
#define L_PRIORITY 30
#define MIN_L_PRIORITY 1
#define MAX_L_PRIORITY 120

#define H_PRIORITY 99
#define MIN_H_PRIORITY 1
#define MAX_H_PRIORITY 120

#define PRIORITY_FORMULA(i,n,low,high) (int)(low*1.0+((high-low)*1.0/(n-1))*i)

/* Defaults section -- End */


#define IS_IN_RANGE(i,min,max) (i >= min && i <= max)

typedef struct thread sos_thread_info;

sos_thread_info idle_thread;
sos_thread_info threads[MAX_NUM_THREADS];

bigtime_t deathTimer;
int lowestPriority;
int highestPriority;
int threadsNumber;
int threadsSpawned;


/* SOS Emulator Addons -- Begin */

typedef struct sos_addon {
	image_id	image;
	char*		name;
} sos_addon;

BList addons;
sos_addon *selectedAddon;
	
struct thread* (*thread_get_current_thread)(void);
void (*scheduler_enqueue_in_run_queue)(struct thread *thread);
void (*scheduler_remove_from_run_queue)(struct thread *thread);
void (*scheduler_reschedule)(void);

sos_addon* find_addon(const char* name)
{
	sos_addon* currAddon;
	int count = addons.CountItems();
	for (int i = 0; i < count; i++) {
		currAddon = (sos_addon *)addons.ItemAt(i);
		if (strcmp(name, currAddon->name) == 0)
			return currAddon;
	}
	return NULL;
}

void load_addon_symbols(image_id &addon_image)
{
	get_image_symbol(addon_image, "thread_get_current_thread",
		B_SYMBOL_TYPE_TEXT, (void**)&thread_get_current_thread);
	get_image_symbol(addon_image, "scheduler_enqueue_in_run_queue",
		B_SYMBOL_TYPE_TEXT, (void**)&scheduler_enqueue_in_run_queue);
	get_image_symbol(addon_image, "scheduler_remove_from_run_queue",
		B_SYMBOL_TYPE_TEXT, (void**)&scheduler_remove_from_run_queue);
	get_image_symbol(addon_image, "scheduler_reschedule",
		B_SYMBOL_TYPE_TEXT, (void**)&scheduler_reschedule);
}

void select_addon(sos_addon *addon)
{
	load_addon_symbols(addon->image);
}

bool load_addons()
{
	image_id	addonId;
	char**		addonName;
   	status_t 	err = B_NO_ERROR; 
    BPath		path; 
	BEntry		entry;
	int32		offset;

	BDirectory	dir("add-ons");

   	//load all add-ons
	while( err == B_NO_ERROR ){
		err = dir.GetNextEntry( (BEntry*)&entry, TRUE );			
		if( entry.InitCheck() != B_NO_ERROR ){
			break;
		}
		if (entry.IsFile())
		if( entry.GetPath(&path) != B_NO_ERROR ){
			printf( "entry.GetPath failed\n" );
		}else{
			BString leaf(path.Leaf());
			offset = leaf.FindLast(".sos_emulator_addon");
			if (offset != B_NO_ERROR) {
				if (leaf.Length() - offset == 19) {
					addonId = load_add_on( path.Path() );
					if ( addonId < 0 ) {
						printf( "load_add_on( %s ) failed\n", path.Path() );
					} else {
						//printf( "load_add_on( %s ) successful!\n", path.Path() );
						if( get_image_symbol( addonId, 
											"add_on_name", 
											B_SYMBOL_TYPE_DATA, 
											(void **)&addonName) ) {
							printf( "get_image_symbol( add_on_name ) failed\n" );
							unload_add_on( addonId );
						} else {
							sos_addon* addon = new sos_addon();
							addon->image = addonId;
							addon->name = *addonName;
							addons.AddItem((void *)addon);
						}
					}
				}
			}
		}
	}
	
	if (addons.IsEmpty()) {
		fprintf(stderr, "Error: No valid SOS Emulator add-ons were found!\n");
		return false;
	}
	return true;
}

/* SOS Emulator Addons -- End */


void sched_func(void*) {
	struct thread *currThread;
	do {
		(*scheduler_reschedule)();
		currThread = (*thread_get_current_thread)();
		currThread->counter += 1;
	} while(1);
}

void print_results_table() {
	int64 sum = 0;
	for (int i=0; i<threadsSpawned; i++) {
		sum += threads[i].counter;
	}
	printf("  Number  Priority            Value    Share (%%)\n");
	printf("--------------------------------------------------\n");
	for (int i=0; i<threadsSpawned; i++) {
		double percent = (threads[i].counter*1.0) / (sum*1.0) * 100;
		printf("%8i %9i %16Lu %12f\n",
			i, threads[i].priority, threads[i].counter, percent);
	}
}

void print_addon_list() {
	int count = addons.CountItems();
	for (int i = 0; i < count; i++) {
		printf("%s", ((sos_addon *)addons.ItemAt(i))->name);
		if (i < count-1)
			printf(" ");
	}
}

// we might get some settings from user
bool get_user_settings(int argc, char** argv) {
	if (argc == 2) {
		if (strcmp("-h",argv[1]) == 0 || strcmp("--help",argv[1]) == 0) {
			printf("usage: %s [<option>] [<time> <t_number> <l_priority> <h_priority> <addon>]\n", argv[0]);
			printf("where:\n");
			printf("      option - '-h' or '--help' to see this help\n");
			printf("               '-l' or '--list' to list all scheduler addons available\n");
			printf("        time - time in milliseconds, between %i and %i\n",
				MIN_DEATH_TIMER_INT32, MAX_DEATH_TIMER_INT32);
			printf("    t_number - number of working threads, between %i and %i\n",
				MIN_NUM_THREADS, MAX_NUM_THREADS);
			printf("  l_priority - the lowest priority possible for a working thread,\n%14s between %i and %i\n",
				"", MIN_L_PRIORITY, MAX_L_PRIORITY);
			printf("  h_priority - the highest priority possible for a working thread,\n%14s between %i and %i\n",
				"", MIN_H_PRIORITY, MAX_H_PRIORITY);
			printf("       addon - is the name of scheduler addon, available names are:\n");
			printf("               ");
			print_addon_list();
			printf("\n");
			printf("defaults: %i %i %i %i %s\n",
				DEATH_TIMER_INT32, NUM_THREADS, L_PRIORITY, H_PRIORITY, selectedAddon->name);
			return false;
		} else if (strcmp("-l",argv[1]) == 0 || strcmp("--list",argv[1]) == 0) {
			print_addon_list();
			printf("\n");
			return false;
		} else {
			fprintf(stderr, "Invalid arguments. Type '%s -h' to get help.\n", argv[0]);
			return false;
		}
	}

	if (argc > 1 && argc != 6) {
		fprintf(stderr, "Invalid arguments. Type '%s -h' to get help.\n", argv[0]);
		return false;
	}

	if (argc == 6) {
		int _deathTimer = atoi(argv[1]);
		int _threadsNumber = atoi(argv[2]);
		int _lowestPriority = atoi(argv[3]);
		int _highestPriority = atoi(argv[4]);
		sos_addon *_selectedAddon = find_addon(argv[5]);
		
		if (IS_IN_RANGE(_deathTimer, MIN_DEATH_TIMER_INT32, MAX_DEATH_TIMER_INT32) &&
			IS_IN_RANGE(_threadsNumber, MIN_NUM_THREADS, MAX_NUM_THREADS) &&			
			IS_IN_RANGE(_lowestPriority, MIN_L_PRIORITY, MAX_L_PRIORITY) &&			
			IS_IN_RANGE(_highestPriority, MIN_H_PRIORITY, MAX_H_PRIORITY) &&		
			_lowestPriority <= _highestPriority && _selectedAddon != NULL) {
			
			deathTimer = _deathTimer * 1000; // conversion to microseconds
			threadsNumber = _threadsNumber;
			lowestPriority = _lowestPriority;
			highestPriority = _highestPriority;
			selectedAddon = _selectedAddon;
		} else {
			fprintf(stderr, "Invalid arguments. Type '%s -h' to get help.\n", argv[5]);
			return false;
		}
	}
	
	return true;
}

int main(int argc, char** argv)
{
	if (!load_addons()) {
		return 1;
	}
	
	deathTimer = DEATH_TIMER_INT32 * 1000; // conversion to microseconds
	threadsNumber = NUM_THREADS;
	lowestPriority = L_PRIORITY;
	highestPriority = H_PRIORITY;
	selectedAddon = (sos_addon *)addons.FirstItem();

	if (!get_user_settings(argc, argv)) {
		return 1;
	}
	select_addon(selectedAddon);
	printf("Settings: work time %Lu microseconds, %i working threads.\n",
		deathTimer, threadsNumber);
	printf("Using %s scheduler add-on.\n\n", selectedAddon->name);

	threadsSpawned = 0;
	for (int i=0; i<threadsNumber; i++) {
		threads[i].priority = PRIORITY_FORMULA(
			i, threadsNumber, lowestPriority, highestPriority
		);
		
		threads[i].counter = 0;
		threadsSpawned++;
	}

	set_thread_priority(find_thread(NULL), B_REAL_TIME_PRIORITY);
		// tell system who is the boss ;)

	idle_thread.priority = 0;
	idle_thread.counter = 0;
	(*scheduler_enqueue_in_run_queue)(&idle_thread);
	
	for (int i=0; i<threadsSpawned; i++) {
		(*scheduler_enqueue_in_run_queue)(&threads[i]);
	}

	thread_id sched_thread_id = spawn_thread((thread_func)sched_func,
			"scheduler emulator", 99, NULL);
	if (sched_thread_id < B_OK) {
		fprintf(stderr, "Can't spawn scheduler thread.");
		return 1;
	}

	resume_thread(sched_thread_id);
	snooze(deathTimer);
	kill_thread(sched_thread_id);

	print_results_table();

	return 0;
}
